Package: terminalgraphics
Type: Package
Title: Graphical Output in Terminals
Version: 0.3.0
Authors@R: 
    c(
      person(
        given = "Jan",
        family = "van der Laan",
        role = c("aut", "cre"),
        email = "r@eoos.dds.nl",
        comment = c(ORCID = "0000-0002-0693-1514")
      ),
      person(
        given = "Doug",
        family = "Kelkhoff",
        role = c("ctb"),
        email = "doug.kelkhoff@gmail.com",
        comment = c(ORCID = "0009-0003-7845-4061")
      )
    )
Description: Defines a graphics device and functions for graphical output in
    terminal emulators that support graphical output. Currently terminals that
    support the Terminal Graphics Protocol
    (<https://sw.kovidgoyal.net/kitty/graphics-protocol/>) and terminal supporting Sixel 
    (<https://en.wikipedia.org/wiki/Sixel>) are supported.
BugReports: https://codeberg.org/djvanderlaan/terminalgraphics/issues
URL: https://codeberg.org/djvanderlaan/terminalgraphics
Depends: R (>= 4.1.0)
Imports: Rcpp, utils, methods, graphics, grDevices, ragg, base64enc
Suggests: magick
LinkingTo: Rcpp
License: GPL-3
SystemRequirements: C++20, A terminal emulator supporting the Terminal
        Graphics Protocol
        (<https://sw.kovidgoyal.net/kitty/graphics-protocol/>) or
        Sixel.
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-05-01 13:53:39 UTC; eoos
Author: Jan van der Laan [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0693-1514>),
  Doug Kelkhoff [ctb] (ORCID: <https://orcid.org/0009-0003-7845-4061>)
Maintainer: Jan van der Laan <r@eoos.dds.nl>
Repository: CRAN
Date/Publication: 2026-05-01 14:10:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-01 15:31:27 UTC; unix
Archs: terminalgraphics.so.dSYM

# Generated by roxygen2: do not edit by hand

export(has_sixel_support)
export(has_tgp_support)
export(is_ghostty)
export(is_kitty)
export(is_nvim)
export(is_tmux)
export(is_vim)
export(kitty_background)
export(kitty_colors)
export(kitty_foreground)
export(kitty_palette)
export(png2sixel)
export(png2terminal)
export(png2tgp)
export(raster2sixel)
export(raster2terminal)
export(raster2tgp)
export(sixel)
export(term_background)
export(term_color_mode)
export(term_dim)
export(term_foreground)
export(term_height)
export(term_palette)
export(term_replot)
export(term_width)
export(tgp)
importFrom(Rcpp,sourceCpp)
importFrom(base64enc,base64encode)
useDynLib(terminalgraphics, .registration = TRUE)

# terminalgraphics

An R-package for graphical output in terminals. Currently supports terminals
that support the [Terminal Graphics
Protocol](https://sw.kovidgoyal.net/kitty/graphics-protocol/) and terminals that
support [Sixel](https://en.wikipedia.org/wiki/Sixel). It implements a graphical
device that will output into the terminal and also some functions for retrieving
information from the terminal such as the colour scheme used, the dimensions of
the terminal and function for displaying png images and image matrices. 

### Example code

Set the `tgp` (Terminal Graphics Protocol) device as the standard device and
instruct the package to use the terminal colors (only supported currently in
Kitty):
```R
options(device = terminalgraphics::tgp, term_col = TRUE)
```

For example, one could use the following lines in `~/.Rprofile` to set the the
`tgp` device as the default device.
```R
if (interactive() && requireNamespace("terminalgraphics", quietly = TRUE) && 
      terminalgraphics::has_tgp_support()) {
  options(device = terminalgraphics::tgp, term_col = TRUE)
}
```
or for Sixel support:
```R
if (interactive() && requireNamespace("terminalgraphics", quietly = TRUE) && 
      terminalgraphics::has_sixel_support()) {
  options(device = terminalgraphics::sixel, term_col = TRUE)
}
```

Creating a plot like, for example:
```R
{
  data(iris)
  plot(iris$Sepal.Width, iris$Sepal.Length, 
    col = term_palette()[iris$Species], 
    pch = 20)
  grid(lty = 2)
  legend("topright", legend = levels(iris$Species), 
    col = term_palette()[1:3], pch = 20)
}
```

Results in:

<img src="kitty_demo.png" width="600" alt="Screenshot of a plot in kitty">


When `term_col = TRUE`, it will try to match the colours of the plot to those
used in the terminal. See [below](#colours) for the limitations of this.


Other functions are:

`png2terminal`, `png2tgp`, `png2sixel`
: display a png file in the terminal.

`raster2terminal`, `raster2tgp`, `raster2sixel`
: display image raster in the terminal.

`term_dim`, `term_width`, `term_height`
: Get the dimensions of the terminal (in pixels and number rows and columns). 

`term_palette` 
: get a palette based on the colours used in the current theme.

`term_color_mode`
: Tries to detect if the terminal is running in dark or light mode. Will return `NA` when this cannot be determined.

`term_foreground` and `term_background` 
: Try to get the foreground and background colours of the terminal.

`has_tgp_support`
: determine if the terminal supports the Terminal Graphics Protocol.

`has_sixel_support`
: determine if the terminal supports Sixel.

`is_kitty`, `is_ghostty`
: is R running in a Kitty or Ghostty terminal.

`term_replot`
: Redraw the current device in the terminal.


### Tips 

Every time the contents of the graphics device change after evaluating an
expression, a new version of the plot is shown in the terminal. Therefore, the
following two lines will result in two plots in the terminal:

```R
plot(rnorm(200))
abline(h = 0)
```

If you want only one plot to appear in the terminal, you should combine the
lines into one expression:

```R
{ 
  plot(rnorm(200))
  abline(h = 0)
}
```

The size and resolution of the plot can be changed using arguments of `tgp` and
`sixel`:
```R
tgp(width = 1000, height = 500)
plot(rnorm(200))
```

It is also possible to set some of these using options:

```R
options(device = terminalgraphics::tgp, term_width = 1000, term_height = 500, 
  term_res = 150, term_col = TRUE)
```

Note that these options are used for *new* devices. Therefore, when there is
already a device open, you should call `tgp`/`sixel` to open a new device or
close open devices using `dev.off` before starting a new plot.

#### Colours

Setting `term_col` to `TRUE` when opening a `sixel` or `tgp` device, these
devices will try to set the colors of the device to match those of the terminal.
This argument can also be set as an option `options(term_col = TRUE)`. However,
currently, only for Kitty can the colours be determined. For some other
terminals it is possible to determine if the terminal is running in dark mode or
light more (see `term_color_mode`) in which case the color scheme will either be
dark or light. Colours can also be set using arguments when opening a `tgp` or
`sixel` device or using options:

```R
options(
  term_col = TRUE,
  term_foreground = "#CCCCCC",
  term_background = "#222222",
  term_palette = grDevices::hcl.colors("Pastel 1", n = 9)
)
```

### False warnings 

The package tries to determine if a terminal supports Sixel or Terminal Graphics
Protocol. Unfortunately, this is far from perfect. This can result in false
warnings when opening a device. These are only issues once and the device will
work. The warnings can be disabled by the options `warned_sixel_support` and
`warned_tgp_support` to `TRUE`. 

# terminalgraphics

An R-package for graphical output in terminals. Currently supports terminals
that support the [Terminal Graphics
Protocol](https://sw.kovidgoyal.net/kitty/graphics-protocol/) and terminals that
support [Sixel](https://en.wikipedia.org/wiki/Sixel). It implements a graphical
device that will output into the terminal and also some functions for retrieving
information from the terminal such as the colour scheme used, the dimensions of
the terminal and function for displaying png images and image matrices. 

### Example code

Set the `tgp` (Terminal Graphics Protocol) device as the standard device and
instruct the package to use the terminal colors (only supported currently in
Kitty):
```R
options(device = terminalgraphics::tgp, term_col = TRUE)
```

For example, one could use the following lines in `~/.Rprofile` to set the the
`tgp` device as the default device.
```R
if (interactive() && requireNamespace("terminalgraphics", quietly = TRUE) && 
      terminalgraphics::has_tgp_support()) {
  options(device = terminalgraphics::tgp, term_col = TRUE)
}
```
or for Sixel support:
```R
if (interactive() && requireNamespace("terminalgraphics", quietly = TRUE) && 
      terminalgraphics::has_sixel_support()) {
  options(device = terminalgraphics::sixel, term_col = TRUE)
}
```

Creating a plot like, for example:
```R
{
  data(iris)
  plot(iris$Sepal.Width, iris$Sepal.Length, 
    col = term_palette()[iris$Species], 
    pch = 20)
  grid(lty = 2)
  legend("topright", legend = levels(iris$Species), 
    col = term_palette()[1:3], pch = 20)
}
```

Results in:

<img src="kitty_demo.png" width="600" alt="Screenshot of a plot in kitty">


When `term_col = TRUE`, it will try to match the colours of the plot to those
used in the terminal. See [below](#colours) for the limitations of this.


Other functions are:

`png2terminal`, `png2tgp`, `png2sixel`
: display a png file in the terminal.

`raster2terminal`, `raster2tgp`, `raster2sixel`
: display image raster in the terminal.

`term_dim`, `term_width`, `term_height`
: Get the dimensions of the terminal (in pixels and number rows and columns). 

`term_palette` 
: get a palette based on the colours used in the current theme.

`term_color_mode`
: Tries to detect if the terminal is running in dark or light mode. Will return `NA` when this cannot be determined.

`term_foreground` and `term_background` 
: Try to get the foreground and background colours of the terminal.

`has_tgp_support`
: determine if the terminal supports the Terminal Graphics Protocol.

`has_sixel_support`
: determine if the terminal supports Sixel.

`is_kitty`, `is_ghostty`
: is R running in a Kitty or Ghostty terminal.

`term_replot`
: Redraw the current device in the terminal.


### Tips 

Every time the contents of the graphics device change after evaluating an
expression, a new version of the plot is shown in the terminal. Therefore, the
following two lines will result in two plots in the terminal:

```R
plot(rnorm(200))
abline(h = 0)
```

If you want only one plot to appear in the terminal, you should combine the
lines into one expression:

```R
{ 
  plot(rnorm(200))
  abline(h = 0)
}
```

The size and resolution of the plot can be changed using arguments of `tgp` and
`sixel`:
```R
tgp(width = 1000, height = 500)
plot(rnorm(200))
```

It is also possible to set some of these using options:

```R
options(device = terminalgraphics::tgp, term_width = 1000, term_height = 500, 
  term_res = 150, term_col = TRUE)
```

Note that these options are used for *new* devices. Therefore, when there is
already a device open, you should call `tgp`/`sixel` to open a new device or
close open devices using `dev.off` before starting a new plot.

#### Colours

Setting `term_col` to `TRUE` when opening a `sixel` or `tgp` device, these
devices will try to set the colors of the device to match those of the terminal.
This argument can also be set as an option `options(term_col = TRUE)`. However,
currently, only for Kitty can the colours be determined. For some other
terminals it is possible to determine if the terminal is running in dark mode or
light more (see `term_color_mode`) in which case the color scheme will either be
dark or light. Colours can also be set using arguments when opening a `tgp` or
`sixel` device or using options:

```R
options(
  term_col = TRUE,
  term_foreground = "#CCCCCC",
  term_background = "#222222",
  term_palette = grDevices::hcl.colors("Pastel 1", n = 9)
)
```

Note that in order to be able to determine all colours in Kitty, the option
```
allow_remote_control yes
```
needs to be set in `~/.config/kitty/kitty.conf`.

### False warnings 

The package tries to determine if a terminal supports Sixel or Terminal Graphics
Protocol. Unfortunately, this is far from perfect. This can result in false
warnings when opening a device. These are only issues once and the device will
work. The warnings can be disabled by the options `warned_sixel_support` and
`warned_tgp_support` to `TRUE`. 

### Tmux

Terminal Graphics Protocol and Sixel are also supported from tmux as long as the
terminal support it. For Terminal Graphics Protocol the `allow-passthrough`
option needs to be set to `on`, which it usually is not by default. This can be
enabled by the following line in `~/.tmux.conf`.

```
set-option -g allow-passthrough on
```

#' Get the color mode of the terminal
#'
#' @details
#' Uses ANSI \code{SI ? 996 n} to determine the color mode. See
#' \url{https://contour-terminal.org/vt-extensions/color-palette-update-notifications/}.
#' 
#' @return
#' Returns \code{NA} when the terminal does not support querying the color 
#' mode. Otherwise returns either 'dark' or 'light'. 
#'
#' @examples
#' term_color_mode()
#'
#' @export
term_color_mode <- function() {
  dsr <- if (is_ghostty()) FALSE else TRUE
  query_color_mode(dsr)
}


query_color_mode <- local({
  color_mode <- NULL
  function(dsr = TRUE) {
    if (is.null(color_mode)) {
      color_mode <<- term_color_mode_rcpp(dsr)
    }
    color_mode
  }
})

#' Dump a PNG image to the terminal 
#'
#' @param filename filename of the PNG image
#'
#' @param method method with which the graphical output is written to the
#' terminal. In case of 'auto', it is first checked if the Terminal Graphics
#' Protocol is supported and if so, this is used. Otherwise, Sixel is used.
#'
#' @return
#' Called for it's side effect of writing the image to the terminal (standard
#' out). Returns \code{NULL} invisibly.
#'
#' @seealso
#' See \code{\link{png2tgp}} for output in Terminal Graphics Protocol.  See
#' \code{\link{png2sixel}} for output in Sixel format. 
#'
#' @export
png2terminal <- function(filename, method = c("auto", "tgp", "sixel")) {
  method <- match.arg(method)
  if (method == "auto") {
     method <- if (has_tgp_support()) "tgp" else "sixel"
  }
  if (method == "tgp") {
    png2tgp(filename)
  } else if (method == "sixel") {
    png2sixel(filename)
  }
}

#' Determine the type of terminal
#'
#' @details
#' These functions try to detect if R is running in a given terminal. 
#'
#' @return
#' Returns \code{TRUE} if R is running in a given terminal
#' and \code{FALSE} otherwise.
#'
#' @examples
#' if (is_kitty()) {
#'   cat("Yeeh, you are running kitty!")
#' }
#' if (is_ghostty()) {
#'   cat("Yeeh, you are running ghostty!")
#' }
#'
#' @rdname is_terminal
#' @export
is_kitty <- function() {
  Sys.getenv("TERM") == "xterm-kitty" || 
    nchar(Sys.getenv("KITTY_PID")) > 0
}

#' @rdname is_terminal
#' @export
is_ghostty <- function() {
  term <- Sys.getenv("TERM")
  term == "xterm-ghostty"
}

#' @rdname is_terminal
#' @export
is_nvim <- function() {
  nvim <- Sys.getenv("NVIM")
  nchar(nvim) > 0
}

#' @rdname is_terminal
#' @export
is_vim <- function() {
  vim <- Sys.getenv("VIM_TERMINAL")
  nchar(vim) > 0
}


#' @rdname is_terminal
#' @export
is_tmux <- function() {
  nchar(Sys.getenv("TMUX")) > 0 ||
    Sys.getenv("TERM_PROGRAM") == "tmux" ||
    grepl("tmux", Sys.getenv("TERM"), fixed = TRUE)
}

#' Get the with of the terminal window in pixels
#'
#' @return
#' An integer with the number of pixels the terminal is wide This value can be
#' zero when the terminal does not support querying the size. Under windows the
#' return value will always be zero as querying the size depends on POSIX
#' support.
#'
#' @seealso
#' \code{\link{term_height}} for the terminal height and \code{\link{term_dim}}
#' for all dimensions including the dimensions in rows and columns.
#'
#' @export
term_width <- function() {
  term_dim()[[1]]
}

#' Get the colors used in the kitty terminal
#'
#' @details
#' To get the background and foreground colors, \code{kitten query-terminal} is
#' called. To get all colors and the palette \code{kitty @get-colors} is called
#' using \code{\link{system}}. However, for the last to work
#' \code{allow_remote_control} needs to be set to \code{yes} in the config
#' file for kitty.
#'
#' @return
#' \code{kitty_colors} returns a data.frame with the colors from the theme used
#' by kitty. \code{kitty_background} returns the background color (character
#' vector with the hex-code). \code{kitty_foreground} returns the foreground
#' color. \code{kitty_palette} returns a vector with the 9 main accent colors
#' from the theme.
#'
#' @seealso
#' \code{\link{term_background}}, \code{\link{term_foreground}},
#' \code{\link{term_palette}} for functions that try to return the colors used
#' in any terminal. When running in kitty, these will call
#' \code{kitty_background} etc.
#'
#' @examples
#' if (is_kitty()) {
#'   cat("The background color is '", kitty_background(), "'.")
#' }
#'
#' @rdname kitty_colors
#' @export
kitty_colors <- function() {
  colors <- getOption("kitty_colors")
  if (is.null(colors)) {
    if (!is_kitty()) {
      stop("Your terminal does not seem to be 'kitty'.")
    }
    colors <- system("kitty @get-colors", intern = TRUE, 
      ignore.stderr = TRUE)
    status <- attr(colors, "status")
    if (!is.null(status) && status != 0) 
      stop(attr(colors, "errmsg"))
    colors <- utils::read.table(
      textConnection(colors),
      comment = "",
      header = FALSE
    )
    names(colors) <- c("name", "value")
    options(kitty_colors = colors)
  }
  colors
}


#' @export
#' @rdname kitty_colors
kitty_background <- function() {
  background <- getOption("kitty_background")
  if (is.null(background)) {
    if (!is_kitty()) {
      stop("Your terminal does not seem to be 'kitty'.")
    }
    background <- system("kitten query-terminal background", intern = TRUE)
    background <- gsub("background: ", "", background)
    options(kitty_background = background)
  }
  background
}

#' @export
#' @rdname kitty_colors
kitty_foreground <- function() {
  foreground <- getOption("kitty_foreground")
  if (is.null(foreground)) {
    if (!is_kitty()) {
      stop("Your terminal does not seem to be 'kitty'.")
    }
    foreground <- system("kitten query-terminal foreground", intern = TRUE)
    foreground <- gsub("foreground: ", "", foreground)
    options(kitty_foreground = foreground)
  }
  foreground
}

#' @export
#' @rdname kitty_colors
kitty_palette <- function() {
  colors <- kitty_colors()
  colors$value[colors$name %in% paste0("color", 1:9)]
}

#' Dump an image raster to the terminal using the Terminal Graphics Protocol
#'
#' @param raster the image 'raster' e.g. the output of \code{\link{as.raster}}.
#'
#' @param compress compress the data before sending to the terminal.
#'
#' @details
#' The Terminal Graphics Protocol is not supported by many Terminal Emulators.
#' The most notable terminal emulator supporting the protocol is Kitty.
#'
#' @return
#' Called for it's side effect of writing the image to the terminal (standard
#' out). Returns \code{NULL} invisibly.
#'
#' @seealso
#' See \code{\link{raster2sixel}} for output in Sixel format. See
#' \code{\link{png2terminal}} for writing a PNG image to the terminal.
#'
#' @importFrom base64enc base64encode
#' @export
raster2tgp <- function(raster, compress = TRUE) {
  warn_tgp_support()

  td <- term_dim()
  support <- query_tgp_support()

  # Convert the raster to raw byte vector
  bytes <- 24
  if (methods::is(raster, "nativeRaster")) {
    r <- bitwAnd(bitwShiftR(raster, 0), 255)
    g <- bitwAnd(bitwShiftR(raster, 8), 255)
    b <- bitwAnd(bitwShiftR(raster, 16), 255)
    a <- bitwAnd(bitwShiftR(raster, 24), 255)
    data <- rbind(r,g,b,a) |> as.vector() |> as.raw()
    bytes <- 32
  } else {
    data <- t(raster) |> grDevices::col2rgb() |> as.vector() |> as.raw() 
  }

  # Calculate image dimensions in units of terminal characters
  cols <- ceiling(ncol(raster) / (td[["x_pixels"]] %/% td[["columns"]]))
  rows <- ceiling(nrow(raster) / (td[["y_pixels"]] %/% td[["rows"]]))

  # Build control character prefix for transmitting image
  control <- character()
  control[["a"]] <- "T" # type of transmission; T = data & image display
  control[["f"]] <- bytes # image format; 24bit RGB
  control[["q"]] <- 2 # suppress terminal response
  control[["s"]] <- ncol(raster) # image width in pixels
  control[["v"]] <- nrow(raster) # image height
  control[["i"]] <- next_tgp_id()

  # If a protocol is required, render as a virtual placeholder
  if (!is.na(support) && nzchar(support)) {
    control[["U"]] <- 1
    control[["c"]] <- cols # number of columns to display image over
    control[["r"]] <- rows # number of rows to display image over
  }

  if (compress) {
    data <- memCompress(data)
    control[["o"]] <- "z" # type of compression
  }

  # Encode
  payloads <- base64enc::base64encode(data, 4096)

  # Send the chunks to the terminal
  for (i in seq_along(payloads)) {
    control[["m"]] <- as.integer(i != length(payloads))
    controlstring <- paste0(names(control), "=", control, collapse = ",")
    out <- paste0("\033_G", controlstring, ";", payloads[i], "\033\\")
    out <- with_tgp_protocol(out)
    cat(out)
  }

  # If a virtual placeholder is used, emit placeholder characters
  if ("U" %in% names(control)) {
    flag <- "\U10EEEE"
    prefix <- paste0("\r\033[38;5;", control[["i"]], "m")
    suffix <- paste0("\033[39m")
    x_cols <- min(as.integer(control[["c"]]), td[["columns"]])
    col_str <- strrep(flag, x_cols - 1L)
    row_str <- tgp_diacritics[seq_len(as.integer(control[["r"]]))]
    row_out <- paste0(prefix, flag, row_str, col_str, suffix)
    cat(paste0(row_out, collapse = "\n"))
  }

  invisible(NULL)
}

next_tgp_id <- local({
  id <- 0L
  function() id <<- id + 1L
})

#' Get the height of the terminal window in pixels
#'
#' @return
#' An integer with the number of pixels the terminal is high. This value can be
#' zero when the terminal does not support querying the size.
#'
#' @seealso
#' \code{\link{term_width}} for the terminal width and \code{\link{term_dim}}
#' for all dimensions including the dimensions in rows and columns.
#'
#' @export
term_height <- function() {
  term_dim()[[2]]
}

#' Determine if terminal supports Sixel
#'
#' @param warn show warnings when the protocol is not supported.
#'
#' @param throw throw an error when we are running in a kitty terminal.
#'
#' @return
#' Returns \code{TRUE} if the current terminal supports the Sixel and
#' \code{FALSE} otherwise. Will also return \code{FALSE} when not running in a
#' terminal.
#'
#' @examples
#' if (has_sixel_support()) {
#'   cat("Yeeh, your terminal supports the sixel!")
#' }
#'
#' @export
has_sixel_support <- function(warn = FALSE, throw = FALSE) {
  if (is_vim() || is_nvim() || is_kitty()) {
    if (warn) 
      warning("Terminal does support sixel")
    return(FALSE)
  }
  # Send query for support
  sup <- query_sixel_support_rcpp(is_tmux())
  if (is.na(sup) || !sup) {
    if (warn) 
      warning("Terminal does not report supporting sixel")
    return(FALSE)
  }
  TRUE
}


warn_sixel_support <- function() {
  warned <- getOption("warned_sixel_support", FALSE)
  if (!warned) {
    if (!has_sixel_support(warn = TRUE)) {
      options("warned_sixel_support" = TRUE)
      warning("This warning is only shown once.")
    }
  }
}

#' Determine if terminal supports Terminal Graphics Protocol
#'
#' @param warn show warnings when the protocol is not supported.
#'
#' @param throw throw an error when we are running in a kitty terminal.
#'
#' @return
#' Returns \code{TRUE} if the current terminal supports the Terminal Graphics
#' Protocol and \code{FALSE} otherwise. Will also return \code{FALSE} when not
#' running in a terminal.
#'
#' @examples
#' if (has_tgp_support()) {
#'   cat("Yeeh, your terminal supports the terminal graphics protocol!")
#' }
#'
#' @export
has_tgp_support <- function(warn = FALSE, throw = FALSE) {
  # Check if the terminal reports its dimensions in pixels; if not we don't
  # want to support it.
  dim <- term_dim()
  if ( length(dim) < 4 || is.na(dim[3]) || dim[3] == 0 || 
       is.na(dim[4]) || dim[4] == 0) {
    if (warn) warning("Terminal does not report its size in pixels.")
    return(FALSE)
  }
  # Send query for support
  supported <- !is.na(query_tgp_support())
  if (!supported) {
    if (warn) {
      warning("Terminal does not report supporting terminal graphics protocol")
      if (tmux() && !tmux_passthrough()) 
        warning("Allow-passthrough is not enabled in tmux.")
    }
    return(FALSE)
  }
  TRUE
}

query_tgp_support <- local({
  # support == NA   no support
  # support == "tmix"   support, but should be escaped
  # support == ""   support
  support <- NULL
  function() {
    if (is.null(support)) {
      if (is_nvim() || is_vim()) {
        # the vim and nvim terminals dont supprt tgp; nvim even
        # hangs
        support <<- NA
      } else if (tmux() && !tmux_passthrough()) {
         # Passthrough is not enabled in tmux
         # terminal graphics protocol is not supported
        support <<- NA
      } else {
        support <<- query_tgp_support_rcpp()
      }
    }
    support
  }
})

warn_tgp_support <- function() {
  warned <- getOption("warned_tgp_support", FALSE)
  if (!warned) {
    if (!has_tgp_support(warn = TRUE)) {
      options("warned_tgp_support" = TRUE)
      warning("This warning is only shown once.")
    }
  }
}


tmux <- function() {
  term <- Sys.getenv("TERM_PROGRAM")
  term == "tmux"
}

tmux_passthrough <- function() {
  term <- Sys.getenv("TERM_PROGRAM")
  if (term != "tmux") return(FALSE)
  sup <- system("tmux show-options -g allow-passthrough", intern = TRUE)
  sup == "allow-passthrough on"
}

#' Dump an image raster to the terminal in Sixel format
#'
#' @param raster the image 'raster' e.g. the output of \code{\link{as.raster}}.
#'
#' @details
#' Sixel is a bitmap format supported by some terminals. See, for example, 
#' \url{https://en.wikipedia.org/wiki/Sixel}.
#'
#' @return
#' Called for it's side effect of writing the image to the terminal (standard
#' out). Returns \code{NULL} invisibly.
#'
#' @seealso
#' See \code{\link{raster2tgp}} for output using Terminal Graphics Protocol.
#'
#' @export
raster2sixel <- function(raster) {
  warn_sixel_support()
  if (!requireNamespace("magick", quietly = TRUE)) 
    stop("The magick package needs to be installed for sixel output.")
  img <- magick::image_read(raster)
  fn <- tempfile()
  on.exit(file.remove(fn))
  magick::image_write(img, format = "sixel", path = fn)
  readChar(fn, nchars = file.size(fn)) |> paste0("\n") |> cat()
}

#' Get the dimensions of the terminal window in pixels
#'
#' @return
#' An integer vector with the width and height of the terminal in pixels
#' (\code{x_pixels} and \code{y_pixels}) and the number of text columns and rows
#' in the terminal window (\code{columns} and \code{rows}).
#'
#' These values can be zero when the terminal does not support querying the
#' size. Some terminals only support querying the number of columns and rows.
#' Under windows the return value will always be zero as querying the size
#' depends on POSIX support.
#'
#' @seealso
#' \code{\link{term_width}} and \code{\link{term_height}} for only obtaining
#' the width and height in pixels respectively.
#'
#' @export
term_dim <- function() {
  dim <- if (is_nvim()) {
    # nvim crashed when trying to query dimensions
    dim <- c(NA_integer_, NA_integer_, NA_integer_, NA_integer_)
  } else {
    screen_dim_cpp()
  }
  names(dim) <- c("x_pixels", "y_pixels", "rows", "columns")
  dim
}

#' Terminal Graphics Protocol Device
#'
#' @param width The width of the image. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param height The height of the image. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param units The units in which 'height' and 'width' are given. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param res The resolution of the image. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param ... passed on to the underlying \code{\link[ragg]{agg_capture}}
#' device.
#'
#' @param term_col Logical value indicating that the foreground and background
#' colors used in the plot should be set to that of the terminal.
#'
#' @param term_bg Logical value indicating that the background color used in
#' the plot should be set to that of the terminal.
#'
#' @param term_fg Logical value indicating that the foreground color used in
#' the plot should be set to that of the terminal.
#'
#' @details
#' The function tries to detect whether the terminal supports Terminal Graphics
#' Protocol. If not, it will issue a warning but still output the image to the
#' terminal. Terminals that to not support Terminal Graphics Protocol will
#' ignore the output. The warning is shown only once. Please file an issue when
#' the terminal does support Terminal Graphics Protocol while
#' \code{\link{has_tgp_support}} returns \code{FALSE}. The warning can be
#' disabled using \code{options(warned_tgp_support = TRUE)}.
#'
#'
#' @return
#' \code{tgp} is called for its side effect of opening a graphics device.
#' Invisibly returns an list with two functions: \code{plot} will plot the
#' current contents of the device in the terminal and \code{update} will plot
#' the current contents of the device in the terminal if the contents have
#' changed since the last plot.
#'
#' \code{\link{term_replot}} will redraw the content of the device in the
#' terminal. In principle \code{\link{term_replot}} is called automatically
#' when the contents of the device changed. This function can be used to force
#' plotting.
#'
#' When \code{term_bg = TRUE} the background color of the graphics device
#' ('\code{bg}') will be set using \code{\link[graphics]{par}}. When
#' \code{term_fg = TRUE} the foreground color ('\code{fg}', '\code{col}',
#' '\code{col.axis}', '\code{col.lab}', '\code{col.main}', and '\code{sub}')
#' will be set using \code{\link[graphics]{par}}.
#'
#' @examples
#' if (has_tgp_support()) {
#'   # Open device
#'   tgp()
#'   plot(rnorm(100))
#'   abline(h = 0, lwd = 2, col = term_palette()[1])
#' }
#'
#' @rdname tgp
#' @export
tgp <- function(
    width = term_default_width(), height = term_default_height(width), 
    units = "px", res = term_default_res(), ...,
    term_col = getOption("term_col", FALSE), term_bg = term_col, term_fg = term_col) {
  warn_tgp_support()
  options <- list(width = width, height = height, units = units, res = res,
    width_set = !missing(width), height_set = !missing(height))
  # Create closure that will handle redrawing
  man <- device_manager(raster2tgp, options, ...)
  cur <- grDevices::dev.cur() |> names()
  if (cur == "null device") stop("Failed to open device")
  if (!exists("devices", devices)) devices$devices <- list()
  devices$devices[[cur]] <- man
  # Set colours
  if (term_bg) { 
    background <- term_background()
    graphics::par(bg = background)
  }
  if (term_fg) { 
    foreground <- term_foreground()
    graphics::par(fg = foreground, col = foreground, 
      col.axis = foreground, col.lab = foreground, 
      col.main = foreground, col.sub = foreground)
  }
  # Add callback handler; this will check after each command if the plot device
  # has changed; and if so, will redraw the plot in the terminal.
  addTaskCallback(term_update, name = "gp")
  setHook("before.plot.new", term_reopen)
  invisible(man)
}


term_default_width <- function() {
  getOption("term_width", max(480, min(1200, term_width(), term_height()/0.8)))
}

term_default_height <- function(width) {
  getOption("term_height", 0.8*width)
}

term_default_res <- function() {
  res <- getOption("term_res", NA)
  if (is.na(res)) {
    dim <- term_dim()
    # number of pixels per row/line = font height
    r <- dim["y_pixels"] / dim["rows"]
    # Default font is 12 points = 12/72 inch so to get the right font size:
    res <- if (is.nan(r) || r == 0) 100 else 0.8 * r * 72 / 12
  }
  res
}

#' Terminal Graphics Protocol Device
#'
#' @param width The width of the image. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param height The height of the image. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param units The units in which 'height' and 'width' are given. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param res The resolution of the image. Passed on to
#' \code{\link[ragg]{agg_capture}}.
#'
#' @param ... passed on to the underlying \code{\link[ragg]{agg_capture}}
#' device.
#'
#' @param term_col Logical value indicating that the foreground and background
#' colors used in the plot should be set to that of the terminal.
#'
#' @param term_bg Logical value indicating that the background color used in
#' the plot should be set to that of the terminal.
#'
#' @param term_fg Logical value indicating that the foreground color used in
#' the plot should be set to that of the terminal.
#'
#' @details
#' The function tries to detect whether the terminal supports Sixel. If not, it
#' will issue a warning but still output the image to the terminal. Terminals
#' that to not support Sixel will ignore the output. The warning is shown only
#' once. Please file an issue when the terminal does support Sixel while
#' \code{\link{has_sixel_support}} returns \code{FALSE}. The warning can be
#' disabled using \code{options(warned_sixel_support = TRUE)}.
#'
#'
#' @return
#' \code{sixel} is called for its side effect of opening a graphics device.
#' Invisibly returns an list with two functions: \code{plot} will plot the
#' current contents of the device in the terminal and \code{update} will plot
#' the current contents of the device in the terminal if the contents have
#' changed since the last plot.
#'
#' \code{\link{term_replot}} will redraw the content of the device in the
#' terminal. In principle \code{\link{term_replot}} is called automatically
#' when the contents of the device changed. This function can be used to force
#' plotting. 
#'
#' When \code{term_bg = TRUE} the background color of the graphics device
#' ('\code{bg}') will be set using \code{\link[graphics]{par}}. When
#' \code{term_fg = TRUE} the foreground color ('\code{fg}', '\code{col}',
#' '\code{col.axis}', '\code{col.lab}', '\code{col.main}', and '\code{sub}')
#' will be set using \code{\link[graphics]{par}}.
#'
#' @examples
#' if (has_sixel_support()) {
#'   # Open device
#'   sixel()
#'   plot(rnorm(100))
#'   abline(h = 0, lwd = 2, col = term_palette()[1])
#' }
#'
#' @rdname sixel
#' @export
sixel <- function(
    width = term_default_width(), height = term_default_height(width), 
    units = "px", res = term_default_res(), ..., 
    term_col = getOption("term_col", FALSE), term_bg = term_col, term_fg = term_col) {
  warn_sixel_support()
  options <- list(width = width, height = height, units = units, res = res,
    width_set = !missing(width), height_set = !missing(height))
  # Create closure that will handle redrawing
  man <- device_manager(raster2sixel, options, ...)
  cur <- grDevices::dev.cur() |> names()
  if (cur == "null device") stop("Failed to open device")
  if (!exists("devices", devices)) devices$devices <- list()
  devices$devices[[cur]] <- man
  # Set colours
  if (term_bg) { 
    background <- term_background()
    graphics::par(bg = background)
  }
  if (term_fg) { 
    foreground <- term_foreground()
    graphics::par(fg = foreground, col = foreground, 
      col.axis = foreground, col.lab = foreground, 
      col.main = foreground, col.sub = foreground)
  }
  # Add callback handler; this will check after each command if the plot device
  # has changed; and if so, will redraw the plot in the terminal.
  addTaskCallback(term_update, name = "gp")
  setHook("before.plot.new", term_reopen)
  invisible(man)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

term_color_mode_rcpp <- function(dsr = TRUE) {
    .Call(`_terminalgraphics_term_color_mode_rcpp`, dsr)
}

query_sixel_support_rcpp <- function(tmux) {
    .Call(`_terminalgraphics_query_sixel_support_rcpp`, tmux)
}

query_tgp_support_rcpp <- function() {
    .Call(`_terminalgraphics_query_tgp_support_rcpp`)
}

screen_dim_cpp <- function() {
    .Call(`_terminalgraphics_screen_dim_cpp`)
}

#' Get the colors used in the terminal
#'
#' @return 
#' \code{term_background} and \code{term_foreground} will return a length 1
#' vector with a color. \code{term_palette} will return a vector of colors.
#'
#' @details
#' Getting the color palette used in the terminal is terminal specific.
#' Currently only the kitty terminal is supported. For other terminals default
#' colors are returned. Unless set using \code{\link{options}} (see below).
#'
#' When the option \code{term_background} is set, that is returned by 
#' \code{term_background()}. The same with the options \code{term_foreground}
#' and \code{term_palette} by \code{term_foreground()} and \code{term_palette()}
#' respectively.
#'
#' @seealso
#' \code{\link{kitty_colors}} for the functions returning the specific colors
#' used in the kitty terminal. Also see \code{\link{term_color_mode}} which is
#' used to determine the color mode.
#'
#' @examples
#' term_background()
#'
#' options(term_background = "black", term_foreground = "white", 
#'   term_palette = grDevices::hcl.colors("Pastel1", n = 9),
#'   term_col = TRUE)
#' # term_col enables automatic use of the colors by tgp() and sixel()
#' 
#' @rdname term_colors
#' @export
term_background <- function() {
  background <- getOption("term_background")
  if (is.null(background)) {
    background <- if (is_kitty()) {
      kitty_background()
    } else if (isTRUE(all.equal(term_color_mode(), "dark"))) {
      "#333333"
    } else {
      "white"
    }
    options(term_background = background)
  }
  background
}

#' @rdname term_colors
#' @export
term_foreground <- function() {
  foreground <- getOption("term_foreground")
  if (is.null(foreground)) {
    foreground <- if (is_kitty()) {
      kitty_foreground()
    } else if (isTRUE(all.equal(term_color_mode(), "dark"))) {
      "white"
    } else {
      "black"
    }
    options(term_foreground = foreground)
  }
  foreground
}

#' @rdname term_colors
#' @export
term_palette <- function() {
  palette <- getOption("term_palette")
  if (is.null(palette) && is_kitty()) {
    try({ 
      palette <- kitty_palette() 
      options(term_palette = palette)
    }, silent = TRUE)
  }
  # fallback
  if (is.null(palette)) {
    if (isTRUE(all.equal(term_color_mode(), "dark"))) {
      palette <- grDevices::hcl.colors("Dark2", n = 9)
    } else {
      palette <- grDevices::hcl.colors("Dark2", n = 9)
    }
    options(term_palette = palette)
  }
  palette
}

source("helpers.R")

library(terminalgraphics)

if (!is_kitty()) {

  expect_equal(term_background(), "white")
  expect_equal(term_foreground(), "black")
  expect_equal(term_palette(), grDevices::hcl.colors("Dark2", n = 9 ))

} 


options(
  term_background = "red", 
  term_foreground = "yellow",
  term_palette = c("pink", "green")
)
expect_equal(term_background(), "red")
expect_equal(term_foreground(), "yellow")
expect_equal(term_palette(), c("pink", "green"))

# Keeps track of changes on a ragg device; returns a list with two functions:
# - plot: will dump the image of the device to the terminal.
# - update: checks is the image has changed and if so, will call plot().
#
#device_manager <- function(ragg_dev, display_fun, dev_open, options) {
device_manager <- function(display_fun, options, ...) {
  ragg_dev <- ragg_open(options, ...)
  # Keep track of the current dimension of the device; when the dimension has
  # changed; the reopen method will open a new device with the updated
  # dimensions.
  dim <- ragg_calc_dim(options)
  # The image; needed to check if there are changes for old versions or ragg
  raster <- ragg_dev()

  reopen <- function() {
    newdim <- ragg_calc_dim(options)
    if (all(newdim == dim)) return();
    dim <<- newdim
    # Keep track of old name of device
    oldname <- grDevices::dev.cur() |> names()
    pos <- which(names(devices$devices) == oldname)
    # Close old device open new device
    grDevices::dev.off()
    # TODO handle ...
    ragg_dev <<- ragg_open(options, ...)
    # Update the list of devices with the new device name
    newname <- grDevices::dev.cur() |> names()
    if (newname == "null device") stop("Failed to open device")
    names(devices$devices)[pos] <- newname
    raster <<- ragg_dev()
  }
  plot <- function(raster_new) {
    if (missing(raster_new)) raster_new <- ragg_dev(native = TRUE)
    raster <<- raster_new
    display_fun(raster)
    cat("\n")
  }
  update <- function() {
    raster_new <- ragg_dev(native = TRUE)
    # ragg > 1.5 can tagg the raster if plot commands have happened
    # this is much faster than comparing the rasters and als much
    # reliable
    if (utils::packageVersion("ragg") > "1.5.0") {
      if (isTRUE(attr(raster_new, "changed"))) {
        plot(raster_new)
      }
    } else {
      if (!isTRUE(all.equal(raster_new, raster))) {
        plot(raster_new)
      }
    }
  }
  list(plot = plot, update = update, reopen = reopen)
}


ragg_open <- function(options, ...) {
  dim <- ragg_calc_dim(options)
  ragg::agg_capture(width = dim[1], height = dim[2], 
    units = options$units, res = options$res, ...)
}

ragg_calc_dim <- function(options) {
  width <- if (options$width_set) options$width else 
    term_default_width()
  height <- if (options$height_set) options$height else 
    term_default_height(width)
  c(width, height)
}


term_update <- function(...) {
  if (!exists("devices", devices)) return()
  cur <- grDevices::dev.cur() |> names()
  man <- devices$devices[[cur]]
  if (!is.null(man)) {
    man$update()
  }
  TRUE
}

term_reopen <- function(...) {
  if (!exists("devices", devices)) return()
  cur <- grDevices::dev.cur() |> names()
  man <- devices$devices[[cur]]
  if (!is.null(man)) {
    man$reopen()
  }
  TRUE
}


#' Replot the current device in the terminal.
#'
#' @return
#' Called for outputting the current contents of a \code{\link{sixel}} or
#' \code{\link{tgp}} device into the terminal.
#'
#' @export
term_replot <- function() {
  if (!exists("devices", devices)) stop("No devices registered")
  cur <- grDevices::dev.cur() |> names()
  man <- devices$devices[[cur]]
  if (!is.null(man)) {
    man$plot()
  }
}

# Send a ANSI-command to the terminal possibly wrapping it in ANSI passthrough 
# codes for a given protocol. This is used to send ANSI codes to the underlying
# terminal when working in tmux.
#
with_tgp_protocol <- function(x) {
  protocol <- query_tgp_support()
  if (nzchar(protocol)) {
    paste0("\033P", protocol, ";", gsub("\033", "\033\033", x), "\033\\")
  } else {
    x
  }
}

#' Dump a PNG image to the terminal using the Terminal Graphics Protocol
#'
#' @param filename filename of the PNG image
#'
#' @details
#' The Terminal Graphics Protocol is not supported by many Terminal Emulators.
#' The most notable terminal emulator supporting the protocol is Kitty.
#'
#' @return
#' Called for it's side effect of writing the image to the terminal (standard
#' out). Returns \code{NULL} invisibly.
#'
#' @importFrom base64enc base64encode
#' @export
png2tgp <- function(filename) {
  warn_tgp_support()
  d <- base64enc::base64encode(filename, 4096)
  for (i in seq_along(d)) {
    out <- "\033_Ga=T,f=100,m=1;"
    end <- "\033\\"
    out <- paste0(out, d[i], end)
    cat(out)
  }
  out <- "\033_Ga=T,f=100,m=0"
  end <- "\033\\"
  out <- paste0(out, end)
  out <- with_tgp_protocol(out)
  cat(out)
  invisible(NULL)
}

expect_equal <- function(x, y, attributes = TRUE) {
  if (is.data.frame(x) && !attributes) 
    return(expect_equal_data.frame(x, y, attributes))
  if (!attributes) attributes(x) <- NULL
  if (!attributes) attributes(y) <- NULL
  stopifnot(isTRUE(all.equal(x, y)))
}

expect_equal_data.frame <- function(x, y, attributes = TRUE) {
  if (!attributes) {
    expect_equal(class(x), class(y))
    expect_equal(names(x), names(y))
    for (col in names(x)) 
      expect_equal(x[[col]], y[[col]], attributes = FALSE)
  } else expect_equal(x, y, attributes)
}

expect_attribute <- function(x, name, value) {
  v <- attr(x, name)
  stopifnot(!is.null(v))
  expect_equal(v, value)
}

expect_error <- function(expr) {
  expect_error.error <- TRUE
  try({
    expr
    expect_error.error <- FALSE
  }, silent = TRUE)
  if (!expect_error.error) stop("Expression did not throw an error.")
}

expect_warning <- function(expr) {
  messages <- list()
  warnings <- list()  
  errors   <- list()
  #tryCatch(
    withCallingHandlers(
      expr, 
      warning = function(w) { 
        warnings <<- append(warnings, list(w))
        invokeRestart("muffleWarning")
      }#,
      #message = function(m) {
      #  messages <<- append(messages, list(m))
      #  invokeRestart("muffleMessage")
      #}
    )#, 
    #error  = function(e) errors <<- append(errors, list(e))
  #)
  stopifnot(length(warnings) > 0)
}
 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster2sixel.R
\name{raster2sixel}
\alias{raster2sixel}
\title{Dump an image raster to the terminal in Sixel format}
\usage{
raster2sixel(raster)
}
\arguments{
\item{raster}{the image 'raster' e.g. the output of \code{\link{as.raster}}.}
}
\value{
Called for it's side effect of writing the image to the terminal (standard
out). Returns \code{NULL} invisibly.
}
\description{
Dump an image raster to the terminal in Sixel format
}
\details{
Sixel is a bitmap format supported by some terminals. See, for example, 
\url{https://en.wikipedia.org/wiki/Sixel}.
}
\seealso{
See \code{\link{raster2tgp}} for output using Terminal Graphics Protocol.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/png2sixel.R
\name{png2sixel}
\alias{png2sixel}
\title{Dump a PNG image to the terminal in Sixel format}
\usage{
png2sixel(filename)
}
\arguments{
\item{filename}{filename of the PNG image}
}
\value{
Called for it's side effect of writing the image to the terminal (standard
out). Returns \code{NULL} invisibly.
}
\description{
Dump a PNG image to the terminal in Sixel format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/png2tgp.R
\name{png2tgp}
\alias{png2tgp}
\title{Dump a PNG image to the terminal using the Terminal Graphics Protocol}
\usage{
png2tgp(filename)
}
\arguments{
\item{filename}{filename of the PNG image}
}
\value{
Called for it's side effect of writing the image to the terminal (standard
out). Returns \code{NULL} invisibly.
}
\description{
Dump a PNG image to the terminal using the Terminal Graphics Protocol
}
\details{
The Terminal Graphics Protocol is not supported by many Terminal Emulators.
The most notable terminal emulator supporting the protocol is Kitty.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster2tgp.R
\name{raster2tgp}
\alias{raster2tgp}
\title{Dump an image raster to the terminal using the Terminal Graphics Protocol}
\usage{
raster2tgp(raster, compress = TRUE)
}
\arguments{
\item{raster}{the image 'raster' e.g. the output of \code{\link{as.raster}}.}

\item{compress}{compress the data before sending to the terminal.}
}
\value{
Called for it's side effect of writing the image to the terminal (standard
out). Returns \code{NULL} invisibly.
}
\description{
Dump an image raster to the terminal using the Terminal Graphics Protocol
}
\details{
The Terminal Graphics Protocol is not supported by many Terminal Emulators.
The most notable terminal emulator supporting the protocol is Kitty.
}
\seealso{
See \code{\link{raster2sixel}} for output in Sixel format. See
\code{\link{png2terminal}} for writing a PNG image to the terminal.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/term_dim.R
\name{term_dim}
\alias{term_dim}
\title{Get the dimensions of the terminal window in pixels}
\usage{
term_dim()
}
\value{
An integer vector with the width and height of the terminal in pixels
(\code{x_pixels} and \code{y_pixels}) and the number of text columns and rows
in the terminal window (\code{columns} and \code{rows}).

These values can be zero when the terminal does not support querying the
size. Some terminals only support querying the number of columns and rows.
Under windows the return value will always be zero as querying the size
depends on POSIX support.
}
\description{
Get the dimensions of the terminal window in pixels
}
\seealso{
\code{\link{term_width}} and \code{\link{term_height}} for only obtaining
the width and height in pixels respectively.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster2terminal.R
\name{raster2terminal}
\alias{raster2terminal}
\title{Dump an image raster to the terminal}
\usage{
raster2terminal(raster, method = c("auto", "tgp", "sixel"), ...)
}
\arguments{
\item{raster}{the image 'raster' e.g. the output of \code{\link{as.raster}}.}

\item{method}{method with which the graphical output is written to the
terminal. In case of 'auto', it is first checked if the Terminal Graphics
Protocol is supported and if so, this is used. Otherwise, Sixel is used.}

\item{...}{passed on to the underlying method such as
\code{\link{raster2tgp}} and \code{\link{raster2sixel}}.}
}
\value{
Called for it's side effect of writing the image to the terminal (standard
out). Returns \code{NULL} invisibly.
}
\description{
Dump an image raster to the terminal
}
\seealso{
See \code{\link{raster2tgp}} for output in Terminal Graphics Protocol.  See
\code{\link{raster2sixel}} for output in Sixel format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/term_height.R
\name{term_height}
\alias{term_height}
\title{Get the height of the terminal window in pixels}
\usage{
term_height()
}
\value{
An integer with the number of pixels the terminal is high. This value can be
zero when the terminal does not support querying the size.
}
\description{
Get the height of the terminal window in pixels
}
\seealso{
\code{\link{term_width}} for the terminal width and \code{\link{term_dim}}
for all dimensions including the dimensions in rows and columns.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/device_manager.R
\name{term_replot}
\alias{term_replot}
\title{Replot the current device in the terminal.}
\usage{
term_replot()
}
\value{
Called for outputting the current contents of a \code{\link{sixel}} or
\code{\link{tgp}} device into the terminal.
}
\description{
Replot the current device in the terminal.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has_tgp_support.R
\name{has_tgp_support}
\alias{has_tgp_support}
\title{Determine if terminal supports Terminal Graphics Protocol}
\usage{
has_tgp_support(warn = FALSE, throw = FALSE)
}
\arguments{
\item{warn}{show warnings when the protocol is not supported.}

\item{throw}{throw an error when we are running in a kitty terminal.}
}
\value{
Returns \code{TRUE} if the current terminal supports the Terminal Graphics
Protocol and \code{FALSE} otherwise. Will also return \code{FALSE} when not
running in a terminal.
}
\description{
Determine if terminal supports Terminal Graphics Protocol
}
\examples{
if (has_tgp_support()) {
  cat("Yeeh, your terminal supports the terminal graphics protocol!")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has_sixel_support.R
\name{has_sixel_support}
\alias{has_sixel_support}
\title{Determine if terminal supports Sixel}
\usage{
has_sixel_support(warn = FALSE, throw = FALSE)
}
\arguments{
\item{warn}{show warnings when the protocol is not supported.}

\item{throw}{throw an error when we are running in a kitty terminal.}
}
\value{
Returns \code{TRUE} if the current terminal supports the Sixel and
\code{FALSE} otherwise. Will also return \code{FALSE} when not running in a
terminal.
}
\description{
Determine if terminal supports Sixel
}
\examples{
if (has_sixel_support()) {
  cat("Yeeh, your terminal supports the sixel!")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/png2terminal.R
\name{png2terminal}
\alias{png2terminal}
\title{Dump a PNG image to the terminal}
\usage{
png2terminal(filename, method = c("auto", "tgp", "sixel"))
}
\arguments{
\item{filename}{filename of the PNG image}

\item{method}{method with which the graphical output is written to the
terminal. In case of 'auto', it is first checked if the Terminal Graphics
Protocol is supported and if so, this is used. Otherwise, Sixel is used.}
}
\value{
Called for it's side effect of writing the image to the terminal (standard
out). Returns \code{NULL} invisibly.
}
\description{
Dump a PNG image to the terminal
}
\seealso{
See \code{\link{png2tgp}} for output in Terminal Graphics Protocol.  See
\code{\link{png2sixel}} for output in Sixel format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kitty_colors.R
\name{kitty_colors}
\alias{kitty_colors}
\alias{kitty_background}
\alias{kitty_foreground}
\alias{kitty_palette}
\title{Get the colors used in the kitty terminal}
\usage{
kitty_colors()

kitty_background()

kitty_foreground()

kitty_palette()
}
\value{
\code{kitty_colors} returns a data.frame with the colors from the theme used
by kitty. \code{kitty_background} returns the background color (character
vector with the hex-code). \code{kitty_foreground} returns the foreground
color. \code{kitty_palette} returns a vector with the 9 main accent colors
from the theme.
}
\description{
Get the colors used in the kitty terminal
}
\details{
To get the background and foreground colors, \code{kitten query-terminal} is
called. To get all colors and the palette \code{kitty @get-colors} is called
using \code{\link{system}}. However, for the last to work
\code{allow_remote_control} needs to be set to \code{yes} in the config
file for kitty.
}
\examples{
if (is_kitty()) {
  cat("The background color is '", kitty_background(), "'.")
}

}
\seealso{
\code{\link{term_background}}, \code{\link{term_foreground}},
\code{\link{term_palette}} for functions that try to return the colors used
in any terminal. When running in kitty, these will call
\code{kitty_background} etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_terminal.R
\name{is_kitty}
\alias{is_kitty}
\alias{is_ghostty}
\alias{is_nvim}
\alias{is_vim}
\alias{is_tmux}
\title{Determine the type of terminal}
\usage{
is_kitty()

is_ghostty()

is_nvim()

is_vim()

is_tmux()
}
\value{
Returns \code{TRUE} if R is running in a given terminal
and \code{FALSE} otherwise.
}
\description{
Determine the type of terminal
}
\details{
These functions try to detect if R is running in a given terminal.
}
\examples{
if (is_kitty()) {
  cat("Yeeh, you are running kitty!")
}
if (is_ghostty()) {
  cat("Yeeh, you are running ghostty!")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/term_width.R
\name{term_width}
\alias{term_width}
\title{Get the with of the terminal window in pixels}
\usage{
term_width()
}
\value{
An integer with the number of pixels the terminal is wide This value can be
zero when the terminal does not support querying the size. Under windows the
return value will always be zero as querying the size depends on POSIX
support.
}
\description{
Get the with of the terminal window in pixels
}
\seealso{
\code{\link{term_height}} for the terminal height and \code{\link{term_dim}}
for all dimensions including the dimensions in rows and columns.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tgp.R
\name{tgp}
\alias{tgp}
\title{Terminal Graphics Protocol Device}
\usage{
tgp(
  width = term_default_width(),
  height = term_default_height(width),
  units = "px",
  res = term_default_res(),
  ...,
  term_col = getOption("term_col", FALSE),
  term_bg = term_col,
  term_fg = term_col
)
}
\arguments{
\item{width}{The width of the image. Passed on to
\code{\link[ragg]{agg_capture}}.}

\item{height}{The height of the image. Passed on to
\code{\link[ragg]{agg_capture}}.}

\item{units}{The units in which 'height' and 'width' are given. Passed on to
\code{\link[ragg]{agg_capture}}.}

\item{res}{The resolution of the image. Passed on to
\code{\link[ragg]{agg_capture}}.}

\item{...}{passed on to the underlying \code{\link[ragg]{agg_capture}}
device.}

\item{term_col}{Logical value indicating that the foreground and background
colors used in the plot should be set to that of the terminal.}

\item{term_bg}{Logical value indicating that the background color used in
the plot should be set to that of the terminal.}

\item{term_fg}{Logical value indicating that the foreground color used in
the plot should be set to that of the terminal.}
}
\value{
\code{tgp} is called for its side effect of opening a graphics device.
Invisibly returns an list with two functions: \code{plot} will plot the
current contents of the device in the terminal and \code{update} will plot
the current contents of the device in the terminal if the contents have
changed since the last plot.

\code{\link{term_replot}} will redraw the content of the device in the
terminal. In principle \code{\link{term_replot}} is called automatically
when the contents of the device changed. This function can be used to force
plotting.

When \code{term_bg = TRUE} the background color of the graphics device
('\code{bg}') will be set using \code{\link[graphics]{par}}. When
\code{term_fg = TRUE} the foreground color ('\code{fg}', '\code{col}',
'\code{col.axis}', '\code{col.lab}', '\code{col.main}', and '\code{sub}')
will be set using \code{\link[graphics]{par}}.
}
\description{
Terminal Graphics Protocol Device
}
\details{
The function tries to detect whether the terminal supports Terminal Graphics
Protocol. If not, it will issue a warning but still output the image to the
terminal. Terminals that to not support Terminal Graphics Protocol will
ignore the output. The warning is shown only once. Please file an issue when
the terminal does support Terminal Graphics Protocol while
\code{\link{has_tgp_support}} returns \code{FALSE}. The warning can be
disabled using \code{options(warned_tgp_support = TRUE)}.
}
\examples{
if (has_tgp_support()) {
  # Open device
  tgp()
  plot(rnorm(100))
  abline(h = 0, lwd = 2, col = term_palette()[1])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/term_color_mode.R
\name{term_color_mode}
\alias{term_color_mode}
\title{Get the color mode of the terminal}
\usage{
term_color_mode()
}
\value{
Returns \code{NA} when the terminal does not support querying the color 
mode. Otherwise returns either 'dark' or 'light'.
}
\description{
Get the color mode of the terminal
}
\details{
Uses ANSI \code{SI ? 996 n} to determine the color mode. See
\url{https://contour-terminal.org/vt-extensions/color-palette-update-notifications/}.
}
\examples{
term_color_mode()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/term_colors.R
\name{term_background}
\alias{term_background}
\alias{term_foreground}
\alias{term_palette}
\title{Get the colors used in the terminal}
\usage{
term_background()

term_foreground()

term_palette()
}
\value{
\code{term_background} and \code{term_foreground} will return a length 1
vector with a color. \code{term_palette} will return a vector of colors.
}
\description{
Get the colors used in the terminal
}
\details{
Getting the color palette used in the terminal is terminal specific.
Currently only the kitty terminal is supported. For other terminals default
colors are returned. Unless set using \code{\link{options}} (see below).

When the option \code{term_background} is set, that is returned by 
\code{term_background()}. The same with the options \code{term_foreground}
and \code{term_palette} by \code{term_foreground()} and \code{term_palette()}
respectively.
}
\examples{
term_background()

options(term_background = "black", term_foreground = "white", 
  term_palette = grDevices::hcl.colors("Pastel1", n = 9),
  term_col = TRUE)
# term_col enables automatic use of the colors by tgp() and sixel()

}
\seealso{
\code{\link{kitty_colors}} for the functions returning the specific colors
used in the kitty terminal. Also see \code{\link{term_color_mode}} which is
used to determine the color mode.
}
